'' cardReader.spin
'
' Read SD cards.
'
' functionality:
' 1. open card
' 2. wait for sector read requests
'
' When a sector read request arrives,
' 1. read the sector dat afrom the card into a cog buffer
' 2. copy the 128 longs from the cog buffer to the passed-in hub address
' 3. set a status flag to indicate completion. 
'
' Let this object know nothing of file systems.
' All filesystem management is done by callers.
'

CON 

  #1                  ' starting at zero can hide bugs as vars often inited to 0                  
  UNKNOWN_CARD_TYPE
  SD_TYPE
                  
  ASSEM_WAITING       ' Codes for managing spin-to-assembly communication  
  ASSEM_SUCCESS       
  ASSEM_FAILURE

  CS_LOW
  CARD_DISABLE
  SEND_COMMAND      
  UNTIL_FE
  SEND_GET_ONE_BYTE   ' These are user-level assembly operations      
  SEND_FFS            
 
  READ_SECTOR  

' Error codes 

  CANNOT_ID_CARD
  BAD_BLOCK_SIZE
  TIMEOUT_ONE
  TIMEOUT_TWO
  ASSEM_FAILURE_ONE
  ASSEM_FAILURE_TWO

         
VAR               

  byte Response[40]
  long ms_rate  
                           
PUB init | val
  
' return 0 on success; non-zero on error

  cardCommandStatus := ASSEM_WAITING

  ms_rate := clkfreq/1_000
   
  if (cardCog > -1)
    return 0

  coginit(7, @cardServer, @cardCommandStatus)

  if (val := card_init) <> 0
    return val
         
  if (val := cardInfo) <> 0
    return val
       
  return 0    

  
PUB readSector(sectorAddress, hubAddress)  

' read card sector identified by "address", copying its 512 bytes into "buffer".
' return nothing

  bufferData := hubAddress
  
  asm(READ_SECTOR, sectorAddress)   
 
        
PUB card_init | type, a, b

  asm(CS_LOW, 0)                          ' CS low during CMD0 selects SPI mode.
  asm(SEND_FFS, 10)                       ' 80 clocks initialize MMC and SD cards
  cardCommand(0)                          ' all cards start with command 0
  asm(CARD_DISABLE, 0)

  a := 0
  b := $FF
            
  cardCommand(55)                  
  cardCommand(41)  
  fileSystemType := SD_TYPE             ' SD init is completed with 41,55 pair  
  repeat while (a++ < 50) and (b > 0)   ' Larger SD cards really can take up to seconds to
    delay(20)                         ' complete their internal initialization, so
    cardCommand(55)                     ' allow plenty of time here.
    delay(20)   
    b := cardCommand(41)
    delay(20)
  
  if b <> 0
    fileSystemType := UNKNOWN_CARD_TYPE
    return CANNOT_ID_CARD
    
  return 0
 

PUB cardCommand(command) | t
'
' Perform card initialization commands.

  asm(CS_LOW, 0)
  asm(SEND_GET_ONE_BYTE, command | $40)
  asm(SEND_GET_ONE_BYTE, 0)
  asm(SEND_GET_ONE_BYTE, 0)
  asm(SEND_GET_ONE_BYTE, 0)
  asm(SEND_GET_ONE_BYTE, 0)
  asm(SEND_GET_ONE_BYTE, $95)
  asm(SEND_GET_ONE_BYTE, $FF)
  asm(SEND_GET_ONE_BYTE, $FF)
  t := cogData
  asm(SEND_GET_ONE_BYTE, $FF)

  return t

  
PUB cardInfo | i, j, READ_BL_LEN,C_SIZE,C_SIZE_MULT,MULT,BLOCK_LEN,BLOCK_NR,MEM, del 

' return 0 on success, non-zero on error
' Read and decode the CSD register.
'
' Algorithm from page 3-15 of document MMCProdMn_rev51.pdf
' "MultiMediaCard Product Manual" rev 5.1, SanDisk corp.

  asm(SEND_COMMAND, 9)
  asm(UNTIL_FE, 0)

  j := 0
    
' This register returns 16 bytes.
' Read more to read 2 byte CRC and two $FF to make sure we're at the end.

  repeat i from 0 to 20
    asm(SEND_GET_ONE_BYTE, $FF)
    Response[j++] := cogData

  asm(CARD_DISABLE, 0)

' READ_BL_LEN: 4 bits from array 83:80; exponent of bytes per block
' value usually "9"

  READ_BL_LEN := $F & Response[5]
 
  if (READ_BL_LEN <> 9)   ' something is wrong unless 9 (512 byte blocks)
    return BAD_BLOCK_SIZE
     
' BLOCK_LEN = 2 (exp read_block_len)
' value usually 512 bytes per block
  BLOCK_LEN := 1 << READ_BL_LEN
  
' C_SIZE_MULT: 3 bits. 2 top bits from array 10, 1 bit from array 11.
' typical values 16M: 2, 64M: 4 
  C_SIZE_MULT := ((  3 & Response[9] ) << 1) + ((128 & Response[10]) << 7)

' MULT is 2 exp (c_size_mult + 2)
' typical values 16 for 16MB card, 64 for 64MB card
  MULT := 1 << (C_SIZE_MULT + 2)

' C_SIZE: 12 bits: 2 lo bits of byte 6, byte 7, 2 hi bits of byte 8
' typical values 16MB: 1963, 64MB: 1959; 256M: 3929?
  C_SIZE := ((3 & Response[6]) << 10) + (Response[7] << 2) + (($C0 & Response[8]) >> 6)

' BLOCK_NR
' typical values 16M: 31424, 64M: 125440, 256M: 250880
  BLOCK_NR := (C_SIZE+1) * MULT
  totalSectors := BLOCK_NR

' bytes: block count times block size
' typical values: 16M: 16089088,  64M: 64225280, 128M? 128450560 
  MEM := BLOCK_NR * BLOCK_LEN
  
  return 0
 
  
PUB asm(command, value) | timeout, status, spin_timeout

' Communicate with remote cog: request services and receive results.
'
' If the cog is busy writing an SD card or something, so lets not try
' anything until it is available and waiting for a command.
' (that timeout value is pure guesstimate)

  spin_timeout := 100000
  
  timeout := 0    
  repeat
    status := cardCommandStatus
  until status == ASSEM_WAITING or timeout++ > spin_timeout
  
  if timeout > spin_timeout
    cardCommandStatus := ASSEM_FAILURE
    return TIMEOUT_ONE

  cogData   := value                           ' First set cogData, because as soon as     
  cardCommandStatus := command                         ' hubCardCommandStatus is set, work will start.
  timeout := 0
    
  repeat
    status := cardCommandStatus
  until status == ASSEM_SUCCESS or status == ASSEM_FAILURE or timeout++ > spin_timeout
  
  if timeout > spin_timeout
    cardCommandStatus := ASSEM_WAITING
    return TIMEOUT_TWO
    
  if cardCommandStatus == ASSEM_SUCCESS                ' success
    cardCommandStatus := ASSEM_WAITING
    return 0
 
  if cardCommandStatus == ASSEM_FAILURE                ' failure of some sort
    cardCommandStatus := ASSEM_WAITING
    return ASSEM_FAILURE_ONE

  cardCommandStatus := ASSEM_WAITING
  
  return ASSEM_FAILURE_TWO


PUB delay(ms) | delt

  delt := (ms * ms_rate) #> 381    
  waitcnt(delt + cnt) 

       
DAT
 
cardCog               long -1
totalSectors          long 0
fileSystemType        long 0

'--------------------------------------------------------------------------------
' These are shared between the hub and the card cog 
        
cardCommandStatus     long 0         
cogData               long 0          
bufferData            long 0
doneFlag              long 0


'-------------------------------------------------------------------  
' Assembly cog program.
'
' This is a server program which listens on a hub (status/data) pair for external commands.

org

cardServer  
 
        mov     cardCommandStatusAddr, par  ' grab hub address of command/status long
        mov     cogDataAddr, par            ' and address of data long.
        add     cogDataAddr, #4
        mov     bufferDataAddr, par         ' Get hub address of buffer data
        add     bufferDataAddr, #8          ' transfer command flag and data.
                                              
        andn    outa, CARD_IN                ' prepare for output to be low
        or      dira, CARD_IN                ' Set pin to output
        
        andn    outa, CARD_CLK               ' prepare for output to be low
        or      dira, CARD_CLK               ' Set pin to output 

        or      outa, CARD_CS                ' prepare for output to be hi
        or      dira, CARD_CS                ' Set pin to output

        andn    dira, CARD_OUT
                                                               
toploop       
        
        rdlong  cogCommandStatusValue, cardCommandStatusAddr    
        
' In these first three states, wait for Spin to do something.
' The normal, rest state for this cog is to detect ASSEM_WAITING and to tightly loop back.
          
        cmp     cogCommandStatusValue, #ASSEM_WAITING wz          
  if_E  jmp     #topLoop
        
        cmp     cogCommandStatusValue, #ASSEM_SUCCESS wz          
  if_E  jmp     #topLoop
        
        cmp     cogCommandStatusValue, #ASSEM_FAILURE wz          
  if_E  jmp     #topLoop
        
' In all the rest of the states, do something, then return success or failure.

        cmp     cogCommandStatusValue, #READ_SECTOR wz          
  if_E  jmp     #readaSector

        cmp     cogCommandStatusValue, #SEND_GET_ONE_BYTE wz          
  if_E  jmp     #sendGetOneByte

        cmp     cogCommandStatusValue, #SEND_FFS wz          
  if_E  jmp     #sendFFs

        cmp     cogCommandStatusValue, #UNTIL_FE wz          
  if_E  jmp     #untilFE

        cmp     cogCommandStatusValue, #SEND_COMMAND wz          
  if_E  jmp     #sendCommand

        cmp     cogCommandStatusValue, #CS_LOW wz          
  if_E  jmp     #csLow
        
        cmp     cogCommandStatusValue, #CARD_DISABLE wz          
  if_E  jmp     #cardDisable

' Fall into badReturn if command not recognized.
'--------------------------------------------------

badReturn
        
        mov     tempHubAddr, #ASSEM_FAILURE         ' error if unknown command received         
        wrlong  tempHubAddr, cardCommandStatusAddr   
        jmp     #topLoop
        
normalReturn
        
        mov     tempHubAddr, #ASSEM_SUCCESS 
        wrlong  tempHubAddr, cardCommandStatusAddr   
        jmp     #topLoop

' end of main loop
'-----------------------------------------------------

' read a sector from the card then write to the hub buffer.

readaSector
        rdlong  cardAddress, cogDataAddr           ' get sector number to read
        rdlong  bufferHubPointer, bufferDataAddr   ' get hub address to write to
                 
        andn    outa, CARD_CS                       ' CS enabled

        call    #sub_readFromAddress
        call    #subuntilFE                        ' loop until card responds

        mov     outerCount, #128                   ' loop 128 times, reading 4 bytes each pass       

:outerLoop        
     
        mov     bytToSend, #$FF
        call    #sendGet
        mov     b3, gotByte
           
        mov     bytToSend, #$FF
        call    #sendGet
        mov     b2, gotByte
   
        mov     bytToSend, #$FF
        call    #sendGet
        mov     b1, gotByte
   
        mov     bytToSend, #$FF
        call    #sendGet
        mov     b0, gotByte

        
        mov     returnLong, b0
        shl     returnLong, #8
        or      returnLong, b1
        shl     returnLong, #8
        or      returnLong, b2
        shl     returnLong, #8
        or      returnLong, b3
     
        wrlong  returnLong, bufferHubPointer       ' write long to hub
        add     bufferHubPointer, #4
                     
        djnz    outerCount, #:outerLoop

        wrbyte  two, bufferHubPointer              ' write the "buffer-ready" status flag
 
        call    #sendFFFF
        or      outa, CARD_CS                      ' CS disabled 
        call    #sendFFFF
           
        jmp     #normalReturn

'--------------------------------------------------------
        
csLow
        andn    outa, CARD_CS  
        jmp     #normalReturn

cardDisable
        call    #sendFFFF
        or      outa, CARD_CS
        call    #sendFFFF  
        jmp     #normalReturn
              
sendCommand        
        rdlong  bytToSend, cogDataAddr
        or      bytToSend, #$40
        call    #sendGet             ' command
        call    #sendFFFF            ' 4 dummy addresses
        call    #sendFFFF
        call    #sendFFFF            ' another
        wrlong  gotByte, cogDataAddr  
        call    #sendFF              ' and one more
        jmp     #normalReturn 

untilFE
        call    #subuntilFE
        jmp     #normalReturn
        
sendGetOneByte                             ' exercise card single-byte IO.
        rdlong  bytToSend, cogDataAddr  
        call    #sendGet
        wrlong  gotByte, cogDataAddr  
        jmp     #normalReturn
         
sendFFs
        rdlong  count, cogDataAddr
:loop
        call    #sendFF
        djnz    count, #:loop  
        jmp     #normalReturn
        
readFromAddress
        rdlong  cardAddress, cogDataAddr
        call    #sub_readFromAddress
        jmp     #normalReturn

'--------------------------------------------------------
' some subroutines

subUntilFE        
        mov     count, #$1FF           ' Typical wait: $F5.  Warning - 9 bits max allowed here. 
:loop
        call    #sendFF
        sub     gotByte, #$FE wz          
  if_E  jmp     #sawFE          
        djnz    count, #:loop
        wrlong  count, cogDataAddr 
        jmp     #badReturn
sawFE
subUntilFE_ret  ret


sendFFFF
        call    #sendFF
        call    #sendFF
sendFFFF_ret    ret


sendFF        
        mov     bytToSend, #$FF
        call    #sendGet       
sendFF_ret      ret

        
sub_readFromAddress        

        mov     bytToSend, #$51               ' read = command 17;  17 or $40 -> $51
        call    #sendGet
        
        mov     bytToSend, cardAddress
        shr     bytToSend, #15                ' (* 512, byte 4) -> (<<9,>>24) -> (>>15) 
        call    #sendGet

        mov     bytToSend, cardAddress
        shr     bytToSend, #7                 ' (* 512, byte 3) -> (<<9,>>16) -> (>>7) 
        call    #sendGet

        mov     bytToSend, cardAddress
        shl     bytToSend, #1                 ' (* 512, byte 2) -> (<<9,>>8) -> (<<1) 
        call    #sendGet

        mov     bytToSend, #0                 ' address byte 1 is always 0
        call    #sendGet
            
        call    #sendFF
        
sub_readFromAddress_ret ret                
        
'-----------------------------------------------------

sendGet 
        mov     sendIndex, #8           ' send 8 bits to card and
        mov     gotByte, #0             ' receive 8 bits in return.
:loop
        shl     gotByte, #1          

        mov     tempHubAddr, bytToSend
        and     tempHubAddr, #$80 wz        
  if_NZ or      outa, CARD_IN            ' when hi bit set, make pin hi 
  if_Z  andn    outa, CARD_IN            ' when hi bit NOT set, make pin lo

        mov     tempHubAddr, ina
        and     tempHubAddr, CARD_OUT wz           
  if_nz or      gotByte, #1             ' when input pin set, then set data bit

        shl     bytToSend, #1

        or      outa, CARD_CLK           ' clock card hi
        andn    outa, CARD_CLK           ' clock card lo 

        djnz    sendIndex, #:loop        
 
sendGet_ret     ret
                
'-------------------------------------------------
' cog variable area           
  
cardCommandStatusAddr    long 0
cogDataAddr              long 0
bufferDataAddr           long 0
bufferHubPointer         long 0
CogCommandStatusValue    long 0

' external device pin IO (and other) masks

CARD_CS  long 1 << 20
CARD_IN  long 1 << 21
CARD_CLK long 1 << 22
CARD_OUT long 1 << 23 

F00mask                 long $F00
FF0mask                 long $FF0
FFFFmask                long $FFFF

cardBufferPtr           long 0    ' address of buffer
cardBufferCounter       long 0    ' longs left in buffer

' general purpose variables
          
bytToSend               long 0    ' data to send to card
gotByte                 long 0    ' data retrieved from card
cardAddress             long 0    ' address used when retrieving full sector
tempHubAddr             long 0    ' used several places: getbyte, etc.
tempSwap                long 0
bytesRead               long 0
RAMaddress              long 0     ' used by RAM access subroutine
RAMdata                 long 0
returnLong              long 0
sendIndex               long 0
count                   long 0
outerCount              long 0
writeIndex              long 0

b3                      long 0
b2                      long 0
b1                      long 0
b0                      long 0

two                     long 2
 
' the end
       