'' FileReader.spin
'
' functionality:
'   List directory contents.
'   Locate file in root directory.

' Note:
'    filesystem parameters and variables are defined in two different places
'
'  1. filesystem parameters    -> DAT area; shared by all instances.
'      These are all fixed once set.
'  2. file-specific variables  -> VAR area; each instance gets its own copies.
'     Some are fixed per file; some vary as file is read.
'
  
CON 

  #1                  ' starting at zero can hide bugs as vars often inited to 0
  FAT16

' Error codes
  #1
  CANNOT_FIND_BOOT_PARTITION
  CARD_INIT_FAILED
  ILLEGAL_FILENAME
  FILE_NOT_FOUND
   
   
OBJ

  c: "CardReader"
             
VAR               

  long fileFirstCluster   ' set upon init then fixed 
  long fileSize           ' also set only once upon init  
  
  long sectorInCluster    ' when tracking a file, this goes 0 to "sectorsPerCluster"
  long currentCluster     ' when tracking a file, this goes from "fileFirstCluster" up.
  long currentFATsector   ' cache management: sector number of cached FAT sector
  long sectorInFile       ' count of most recently read sector of this file
  long sectorsInFile      ' total count of sectors in file; filesize/512, rounded up.

  long firstSector, lastSector
      
  byte dirEntry[32]       ' used when analyzing directory entries
    
  byte buffer[512]        ' sector cache used during initial directory read, then to cache FATs.
  byte bufferStatus
  
                           
PUB init(filename) | val
  
' return 0 on success; error code on failure
  
  if (val := c.init) <> 0                      ' initialize the card reader
    return val
             
  if (val := filesystem_init) <> 0             ' initialize the filesystem
    return val
    
  if (val := getFirstCluster(filename)) <> 0   ' locate the requested file
    return val 

  sectorInFile := sectorsInFile := 0

  repeat
    sectorsInFile++                            ' count sectors in file
  while (sectorsInFile << 9) < fileSize

  firstSector := lastSector := -1
            
  return 0

 
PUB getFileSize

  return fileSize   

  
PUB getFileFirstSector(addr) 

  sectorInFile := 0
  sectorInCluster := -1
  currentCluster := fileFirstCluster
  
  firstSector := cardDataStart + ((fileFirstCluster-2) * sectorsPerCluster)

  c.readSector(firstSector, addr)

  repeat
  until bufferStatus > 0

  bufferStatus := 1
        
  return firstSector


PUB resetFile

  sectorInFile := 0
  sectorInCluster := -1
  currentCluster := fileFirstCluster

  
PUB getNextSectorBlocking(addr) | s

' Read a sector from the card. Do not return until the status flag says ready.

  if sectorInFile < sectorsInFile '(sectorsInFile - 1)
    sectorInFile++
  else
    if lastSector == -1
      lastSector := s    
      
    sectorInFile := 0
    sectorInCluster := -1
    currentCluster := fileFirstCluster
 
  s :=  calcNextSector  
  c.readSector(s, addr)

  repeat                         ' block until card has finished 
  until bufferStatus > 0         ' retrieving data
      
  return s 

  
PRI calcNextSector | neededFATsector, index

' return sector number of next sector of file.

  sectorInCluster++
  
' 1. Is next sector available from the current cluster?

  if sectorInCluster < sectorsPerCluster
    return cardDataStart + ((currentCluster-2) * sectorsPerCluster) + sectorInCluster
  
' 2. Get FAT sector number containing next cluster. 

  neededFATsector := fatStart + (currentCluster >> 8) 

' 3. if FAT sector is not in the cache then get it.

  if currentFATsector <> neededFATsector
    buffer[512] := 0
    c.readSector(neededFATsector, @buffer)    
    currentFATsector := neededFATsector 
    repeat 
    until bufferStatus > 0
     
' 4. Get next cluster from the FAT
      
  index :=  ($FF & currentCluster) << 1    
  currentCluster := buffer[index] + (buffer[index+1] << 8)
  sectorInCluster := 0
  return cardDataStart + ((currentCluster-2) * sectorsPerCluster) + sectorInCluster   


PRI getFirstCluster(fileName) | sector, entry, index, i, j

' scan every root dir entry looking for this filename                    
' return ERROR if not found, cluster number otherwise
  
  entry := fileSize := fileFirstCluster := sectorInCluster := sectorInFile := 0
        
  if breakUpFilename(fileName) == false   ' break into prefix and suffix
    return ILLEGAL_FILENAME
  
  sector := dirStart
  
  repeat 
    c.readSector(sector++, @buffer)
    index := 0
    
    repeat j from 0 to 15                 ' There are 16 directory entries per sector
      entry++
      
      repeat i from 0 to 31               ' There are 32 bytes per directory entry
        dirEntry[i] := buffer[index++]

      if testDirEntryForFilename
        fileSize := dirEntry[$1C] + (dirEntry[$1D] << 8) + (dirEntry[$1E] << 16) + (dirEntry[$1F] << 24)
        fileFirstCluster := currentCluster := dirEntry[$1A] + (dirEntry[$1B] << 8)       
        return 0 

  while entry < numberOfRootDirEntries-1

  if (fileFirstCluster == 0)
    return FILE_NOT_FOUND 

  
PRI breakUpFilename(fileName) | i, dotpos

' Break name because DOS allows trailing spaces to exist, but are not
' considered part of the name; ie, "name.txt" == "name .txt" == "name    .txt"
                                                        
' fileName: 1 to 8 prefix characters, a dot, exactly 3 suffix characters.    

  if strsize(fileName) > 12
    return false
    
  dotpos := 0                          
  repeat strsize(fileName) - 1                       ' Note: "fileName" used WITHOUT the @,
    byte[@prefix][dotpos] := byte[fileName][dotpos]  ' but prefix REQUIRES @
    if byte[fileName][dotpos] == $2E 
      quit 
    dotpos++
  byte[@prefix][dotpos] := 0
          
  if dotpos < 1
    return false
    
  byte[@suffix][0] := byte[fileName][dotpos+1]
  byte[@suffix][1] := byte[fileName][dotpos+2]
  byte[@suffix][2] := byte[fileName][dotpos+3]
  byte[@suffix][3] := 0
 
  return true

   
PRI testDirEntryForFilename | i
  
' Check that the first non-space characters match
          
  repeat i from 0 to strsize(@prefix) - 1
    if byte[@dirEntry][i] <> byte[@prefix][i] 
      return false

  return true
 
               
PRI filesystem_init |  bytesPerSector, reserved

  currentFATsector := 0
 
  if !find_boot_partition
    return CANNOT_FIND_BOOT_PARTITION

  entriesPerFAT := 256 
  
' buffer should contain the filesystem boot sector.
' read a bunch of filesystem parameters.

  bytesPerSector         := buffer[$B]  + (buffer[$C] << 8)
  sectorsPerCluster      := buffer[$D]
  reserved               := buffer[$E]  + (buffer[$F] << 8)
  numFATs                := buffer[$10]  
  numberOfRootDirEntries := buffer[$11] + (buffer[$12] << 8) 
  totalFsSectors         := buffer[$13] + (buffer[$14] << 8)

  if (totalFsSectors == 0)
    totalFsSectors       := buffer[$20] + (buffer[$21] << 8) +(buffer[$22] << 16) +(buffer[$23] << 24)
  
  sectorsPerFAT          := buffer[$16] + (buffer[$17] << 8) 

  FATStart := partitionStart + reserved       
  dirStart := FATStart + (sectorsPerFAT * numFATs)
  cardDataStart := dirStart + (numberOfRootDirEntries >> 4)  ' (32 chars/entry) / 512 chars/sector) -> %16 -> >>4 

' force "getNextSector" to reset itself on first use and get the first file sector
  
  sectorInFile := sectorsInFile
    
  return 0

 
PRI find_boot_partition 

' Find the first sector of a filesystem. Some cards contain nothing but a
' filesystem, where sector 0 of the card is the boot sector: 
  
  if is_boot_sector(0)
    return true

' If sector 0 is not a boot sector then it should contain partition table entries.    
' DOS may have up to 4 partitions, with filesystems in any or all of them.
'
' Each of 4 partition table addresses are 16 bytes starting at address $1BE, $1CE, $1DE, $1EE
' Four byte address of first sector in each partition is 8 bytes into each table. 

  if is_boot_sector(buffer[$1C6]+(buffer[$1C7]<<8)+(buffer[$1C8]<<16)+(buffer[$1C9]<<24)) 
    return true
  if is_boot_sector(buffer[$1D6]+(buffer[$1D7]<<8)+(buffer[$1D8]<<16)+(buffer[$1D9]<<24)) 
    return true
  if is_boot_sector(buffer[$1E6]+(buffer[$1E7]<<8)+(buffer[$1E8]<<16)+(buffer[$1E9]<<24)) 
    return true
  if is_boot_sector(buffer[$1F6]+(buffer[$1F7]<<8)+(buffer[$1F8]<<16)+(buffer[$1F9]<<24)) 
    return true                     

  return false

  
PRI is_boot_sector(start)

  partitionStart := start
              
  c.readSector(partitionStart, @buffer) 
  
  if (buffer[$1FE] <> $55) or (buffer[$1FF] <> $AA)  ' 55AA FAT signature
    return false    
'                    F              A                      T                                        
  if buffer[$36] == $46 and buffer[$37] == $41 and buffer[$38] == $54
'                      1                      6 
    if buffer[$39] == $31 and buffer[$3A] == $36
      fileSystemType := FAT16       
      return true
          
  return false

     
DAT
 
cardCog               long -1

'--------------------------------------------------------------------------------
' filesystem parameters - shared by all instances of this object

FatStart               long 0
sectorsPerCluster      long 0 
cardDataStart          long 0  
entriesPerFAT          long 0       
numberOfRootDirEntries long 0
totalFsSectors         long 0
sectorsPerFAT          long 0
partitionStart         long 0
dirStart               long 0
numFATs                long 0
fileSystemType         long 0

prefix byte "yyyyzzzz",0
suffix byte "xxx",0
   
'-------------------------------------------------------------------  